////////////////////////////////////////
//
// Free for private and commercial use.
//
////////////////////////////////////////

#include <zneo.h>
#include "zneo2.h"

#include "application.h"
#include "sio.h"
#include "hyper.h"
#include "flash.h"
#include "loadhex.h"

////////////////////////////////////////

static int isxdigit(int c)
{
	return (c >= '0' && c <= '9') || (c >= 'A' && c <= 'F') || (c >= 'a' && c <= 'f') ? 1 : 0;
}

// This will convert an ascii hex character to it's binary value return -1 on error
static int AsciiToHex(int c)
{
	if (c >= '0' && c <= '9')
		return c - '0';
	else if (c >= 'A' && c <= 'F')
		return c - 'A' + 10;
	else if (c >= 'a' && c <= 'f')
		return c - 'a' + 10;
	else
		return -1;
}

////////////////////////////////////////

extern erom BOOTLOADER_LOW;		// Def as additional linker directive; identifies the lowest address of the Boot Loader
extern erom BOOTLOADER_HIGH;	// Def as additional linker directive; identifies the highest address of the Boot Loader

extern erom RAM_LOW;			// Def as additional linker directive; identifies the lowest address of the RAM
extern erom RAM_HIGH;			// Def as additional linker directive; identifies the highest address of the RAM

extern erom ERAM_LOW;			// Def as additional linker directive; identifies the lowest address of the ERAM
extern erom ERAM_HIGH;			// Def as additional linker directive; identifies the highest address of the ERAM

#define BACKUP_BASE_ADDR	0x0800

int8_t LoadHexMain(uint32_t* puLinesInFile, uint32_t* puTotalBytes, uint32_t* puConflictAddr, uint32_t* puNotWritable, uint32_t* puDownLoadErrors)
{
	int8_t cState = WAITFOR_RECORDBEGIN;
	char cChar, cPrevChar = '\0';
	uint8_t ucIsBootloader = 0, ucBytesExpected = 0, ucRecType = 0, ucDataByte = 0, ucSumCheckValue = 0, ucSumByte = 0;
	uint16_t usHexCharCtr = 0;
	uint32_t uLinAddr = 0, uOffset = 0, uDataAddr;
	int32_t iCtr;

	for (uDataAddr = (uint32_t)&BOOTLOADER_LOW; uDataAddr; EraseFlash(uDataAddr -= PAGE_SIZE)) ;	// erase flash
	WriteFlash(RESET, (uint32_t)&c_startup >> 24);		// restore flash loader entry address
	WriteFlash(RESET + 1, (uint32_t)&c_startup >> 16);
	WriteFlash(RESET + 2, (uint32_t)&c_startup >> 8);
	WriteFlash(RESET + 3, (uint32_t)&c_startup);

	do
	{
		cChar = getch();

		if (cChar == ESC || cChar == CTRL_C)
		{	// User terminated load
			cState = LOADHEX_USERABORT;
			break;
		}
		else if (cChar == ':' && cState != WAITFOR_RECORDBEGIN)
		{
			if (cPrevChar == '\n')
				++*puLinesInFile;
			else
				++*puDownLoadErrors;

			cState = WAITFOR_RECORDBEGIN;	// wait for ':'
		}

		// Process the char depending on the state of download
		switch (cState)
		{
		case WAITFOR_RECORDBEGIN:
			switch(cChar)
			{
			case ' ':						// Ignore white space before record begins
			case '\t':
				break;
			case '\n':
				if (cPrevChar == '\r')
					break;
			case '\r':						// keep track of No.of Lines in hex file
				++*puLinesInFile;
				break;
			case ':':
				usHexCharCtr = 0;
				ucSumCheckValue = 0;		// Init the byte checksum to 0
				ucBytesExpected = 0;
				cState = RECORDLENGTH;
				break;
			case '*':						// Allow for * delimited lines
				cState = COMMENTFIELD;
				break;
			default:						// We are recving out of place characters/ damaged record stream
				++*puDownLoadErrors;
				cState = COMMENTFIELD;
				break;
			}
			break;
		case COMMENTFIELD:					// Wait till resynchronization opportunity
			if (cChar == '\r' || cChar == '\n')
				cState = WAITFOR_RECORDBEGIN;
			break;
		case RECORDLENGTH:					// a 2 character record lenght field
			if (!isxdigit(cChar))
			{
				cState = COMMENTFIELD;
				break;
			}

			ucBytesExpected = (ucBytesExpected << 4) | AsciiToHex(cChar);

			if (++usHexCharCtr & 1)
				break;

			ucSumCheckValue += ucBytesExpected;
			cState = OFFSET;
			break;
		case OFFSET:						// a 4 character offset field
			if (!isxdigit(cChar))
			{
				cState = COMMENTFIELD;
				break;
			}

			uOffset = ((uOffset << 4) | AsciiToHex(cChar)) & 0xFFFF;

			if (++usHexCharCtr & 1)
				break;

			ucSumCheckValue += uOffset & 0xFF;

			if (usHexCharCtr < 6)
				break;

			cState = RECORDTYPE;
			break;
		case RECORDTYPE:					// a 2 character record type field
			if (!isxdigit(cChar))
			{
				cState = COMMENTFIELD;
				break;
			}

			ucRecType = (ucRecType << 4) | AsciiToHex(cChar);

			if (++usHexCharCtr & 1)
				break;

			ucSumCheckValue += ucRecType;

			switch (ucRecType)
			{
			case DATA_RECORD:
				cState = ucBytesExpected > 0 ? DATAFIELD : SUMCHECK;
				break;
			case EOF_RECORD:
				cState = SUMCHECK;
				break;
			case EXT_LIN_ADDR_RECORD:
				cState = EXT_LIN_ADDR_FIELD;
				break;
			default:
				++*puDownLoadErrors;
				cState = COMMENTFIELD;
				break;
			}
			break;
		case EXT_LIN_ADDR_FIELD:			// a 4 character extended linear address field
			if (!isxdigit(cChar))
			{
				++*puDownLoadErrors;
				cState = COMMENTFIELD;
				break;
			}

			uLinAddr = (uLinAddr << 4) | AsciiToHex(cChar);

			if (++usHexCharCtr & 1)
				break;

			ucSumCheckValue += uLinAddr & 0xFF;

			if (--ucBytesExpected)
				break;

			uLinAddr <<= 16;
			cState = SUMCHECK;
			break;
		case DATAFIELD:						// the data field
			if (!isxdigit(cChar))
			{
				++*puDownLoadErrors;
				cState = COMMENTFIELD;
				break;
			}

			ucDataByte = (ucDataByte << 4) | AsciiToHex(cChar);

			if (++usHexCharCtr & 1)
				break;

			ucSumCheckValue += ucDataByte;
			uDataAddr = uLinAddr + uOffset++;

			if ((uDataAddr >= (uint32_t)&RAM_LOW && uDataAddr <= (uint32_t)&RAM_HIGH && &RAM_LOW != &RAM_HIGH) ||
				(uDataAddr >= (uint32_t)&ERAM_LOW && uDataAddr <= (uint32_t)&ERAM_HIGH && &ERAM_LOW != &ERAM_HIGH))
			{	// byte inside internal or external RAM
				*(uint8_t*)uDataAddr = ucDataByte;
			}
			else if (uDataAddr >= RESET && uDataAddr <= RESET + 3)
			{
				// is this a Update of the Boot Loader self ?
				if (uDataAddr == RESET && ucDataByte == (uint8_t)((uint32_t)&c_startup >> 24))
					ucIsBootloader |= _BV(3);
				else if (uDataAddr == RESET + 1 && ucDataByte == (uint8_t)((uint32_t)&c_startup >> 16))
					ucIsBootloader |= _BV(2);
				else if (uDataAddr == RESET + 2 && ucDataByte == (uint8_t)((uint32_t)&c_startup >> 8))
					ucIsBootloader |= _BV(1);
				else if (uDataAddr == RESET + 3 && ucDataByte == (uint8_t)((uint32_t)&c_startup))
					ucIsBootloader |= _BV(0);

				WriteFlash((uint32_t)PTR_USER_ENTRY + (uDataAddr - RESET), ucDataByte);
			}
			else if (uDataAddr >= (uint32_t)PTR_USER_ENTRY && uDataAddr <= (uint32_t)PTR_USER_ENTRY + 3)
				;
			else if (uDataAddr >= (uint32_t)&BOOTLOADER_LOW && uDataAddr <= (uint32_t)&BOOTLOADER_HIGH)
			{
				if (ucIsBootloader == (_BV(3) | _BV(2) | _BV(1) | _BV(0)))
				{	// write new boot loader code to backup area
					WriteFlash(BACKUP_BASE_ADDR + (uDataAddr - (uint32_t)&BOOTLOADER_LOW), ucDataByte);
				}
				else
				{	// user application code overwrites the boot loader
					if (!*puConflictAddr)
						*puConflictAddr = uDataAddr;

					*puNotWritable += ucBytesExpected;
					putch('a');
					cState = COMMENTFIELD;
					break;
				}
			}
			else
				WriteFlash(uDataAddr, ucDataByte);	// free Flash Memory

			++*puTotalBytes;

			if (!--ucBytesExpected)
				cState = SUMCHECK;
			break;
		case SUMCHECK:
			if (!isxdigit(cChar))
			{		 
				cState = COMMENTFIELD;
				break;
			}

			ucSumByte = (ucSumByte << 4) | AsciiToHex(cChar);

			if (++usHexCharCtr & 1)
				break;

			putch((ucSumCheckValue += ucSumByte) ? (++*puDownLoadErrors, 'd') : '.');

			cState = ucRecType == EOF_RECORD ? (++*puLinesInFile, LOADHEX_OK) : WAITFOR_RECORDBEGIN;
			break;
		default:
			++*puDownLoadErrors;
			cState = COMMENTFIELD;
			break;
		}

		cPrevChar = cChar;
	}
	while (cState > LOADHEX_OK);

	iCtr = 0;
	do
	{	// receive and reject subsequent characters
		if (kbhit())
		{
			getch();
			iCtr = 0;
		}
	}
	while (++iCtr < 20000);

	if (cState == LOADHEX_OK && !*puDownLoadErrors && ucIsBootloader == (_BV(3) | _BV(2) | _BV(1) | _BV(0)))	// boot loader update
		HyperFlash((uint32_t)&BOOTLOADER_LOW, BACKUP_BASE_ADDR, (uint32_t)&BOOTLOADER_HIGH + 1 - (uint32_t)&BOOTLOADER_LOW);

	return cState;
}
