////////////////////////////////////////
//
// Free for private and commercial use.
//
////////////////////////////////////////

#include <zneo.h>
#include "zneo2.h"

#include "application.h"
#include "sio.h"
#include "flash.h"
#include "loadhex.h"

////////////////////////////////////////

static char* UintToAscii(uint32_t uiVal, char* pszDst, int iRadix)
{
	char cChar, *pszTmp, *pszStart = pszDst;

	do
	{	// generate digits in reverse order
		int32_t t = uiVal % iRadix;

		if (t > 9)
			*(pszDst++) = (t - 10) + 'A';
		else
			*(pszDst++) = t + '0';
	}
	while (uiVal /= iRadix);

	for (pszTmp = pszDst; --pszTmp > pszStart; )
	{
		cChar = *pszTmp;
		*pszTmp = *pszStart;
		*pszStart++ = cChar;
	}

	return pszDst;
}

static int Puts(char erom* str)
{
	for( ; *str; ++str)
		putch(*str);
	putch('\n');
	putch('\r');
	return(0);
}

static int PutsEx(char erom* pszText, uint32_t uiVal)
{
	char szBuf[100];
	char* pszDst;

	for (pszDst = szBuf; *pszText; )
	{
		if (*pszText != '%')
		{
			*pszDst++ = *pszText++;
			continue;
		}

		switch (++pszText, *pszText++)
		{
		case 'c':
			*pszDst++ = (char)uiVal;
			break;
		case 'd':
			pszDst = UintToAscii(uiVal, pszDst, 10);
			break;
		case 'X':
			pszDst = UintToAscii(uiVal, pszDst, 16);
			break;
		}
	}

	*pszDst = '\0';
	Puts(szBuf);
	return(0);
}

static void ExecuteUserApp(int port)
{
	uint8_t ucOscctl;

	// Reset Flash Frequency register
	InitFlash(0);

	// Reset UARTx
	switch (port)
	{
	case UART_USED:
		U0CTL0 = 0x00;
		U0CTL1 = 0x00;
		PAAFL &= ~0x30;
		U0BR = 0xFFFF;
		break;
	case _UART1:
		U1CTL0 = 0x00;
		U1CTL1 = 0x00;
		PDAFL &= ~0x30;
		U1BR = 0xFFFF;
		break;
	}

	// Reset Oscillator Control Register
	ucOscctl = OSCCTL;
	OSCCTL = OSCCTL_UNLOCK1;
	OSCCTL = OSCCTL_UNLOCK2;
	OSCCTL = ucOscctl | OSCCTL_WDTEN;	// WDT Oscillator Enable

	// Executing User Application
	((void(*)(void))*PTR_USER_ENTRY)();
}

////////////////////////////////////////

void main(void)
{
	uint8_t ucOscctl;
	int32_t iCtr;
	char cChar;

	// Voltage stabilization at the FT232R (discard unwanted ghosts)
	for (iCtr = 500000; iCtr--; ) ;

	if (*PTR_USER_ENTRY != 0xFFFFFFFF && *PTR_USER_ENTRY != (uint32_t)&c_startup)
	{	// User Program code is loadet and this is not a boot loader update
		if (!(OPTION0 & OPTION0_FWP) ||	// Programming, Page Erase, and Mass Erase through user code is disabled
			!(UART_USED == _UART0 ? (PAIN & _BV(4)) : (PDIN & _BV(4))))	// USB Connection not present
			ExecuteUserApp(-1);
	}

	// WDT Oscillator Disable
	ucOscctl = OSCCTL;
	OSCCTL = OSCCTL_UNLOCK1;
	OSCCTL = OSCCTL_UNLOCK2;
	OSCCTL = ucOscctl & ~OSCCTL_WDTEN;

	// init UART
	init_uart(UART_USED, _DEFFREQ, UART_BAUD);

	// init Flash Frequency register
	InitFlash(_DEFFREQ);

	if (*PTR_USER_ENTRY != (uint32_t)&c_startup)
	{	// not boot loader update
		getch();			// wait for second transmission of the PC
		for (iCtr = 32; iCtr--; putch('\n')) ;	// clear screen
	}

	Puts("\n\n " APP_NAME " Boot Loader " APP_VERSION);		// send signs of life

	iCtr = 0;
	do
	{	// reject redundant characters
		if (kbhit())
		{
			getch();
			iCtr = 0;
		}
	}
	while (++iCtr < 20000);

	for (;;)
	{
		Puts("\n\t========= Help =========");
		Puts(" D   Download Hex file to " APP_NAME);
		if (*PTR_USER_ENTRY != 0xFFFFFFFF && *PTR_USER_ENTRY != (uint32_t)&c_startup)
			PutsEx(" G   Execute " APP_NAME " Program code from %Xh\n", *PTR_USER_ENTRY);
		else
			putch('\n');
		putch('>');
		while ((cChar = getch()) < ' ' || cChar > '~') ;
		putch(cChar);
		putch('\n');

		switch (cChar)
		{
		case 'd':
		case 'D':
		{
			uint32_t uLinesInFile = 0, uTotalBytes = 0, uConflictAddr = 0, uNotWritable = 0, uDownLoadErrors = 0;
			int8_t cState;

			Puts("\n Please transmit Hex file as Text file.."); 

			cState = LoadHexMain(&uLinesInFile, &uTotalBytes, &uConflictAddr, &uNotWritable, &uDownLoadErrors);

			if (cState == LOADHEX_USERABORT)
				Puts("\n Download aborted by user");
			else 
			{
				Puts("\n Download completed");

				Puts("\n\t=== Download Summary ===");
				PutsEx(" Start/Reset Address\t\t: %Xh", *PTR_USER_ENTRY);
				PutsEx(" No. of Bytes Flashed\t\t: %d", uTotalBytes);
				PutsEx(" No. of Bytes Not Writable\t: %d", uNotWritable);
				if (uConflictAddr)
					PutsEx(" First Conflict Address at\t: %Xh", uConflictAddr);
				PutsEx(" No. of Hex File Lines\t\t: %d", uLinesInFile);
				PutsEx(" No. of Hex File Download Errors: %d", uDownLoadErrors);
			}
			break;
		}
		case 'g':
		case 'G':
			if (*PTR_USER_ENTRY == 0xFFFFFFFF || *PTR_USER_ENTRY == (uint32_t)&c_startup)
				break;		// User Program code not loadet

			PutsEx("\n Executing " APP_NAME " Program code from %Xh\n", *PTR_USER_ENTRY);
			for (iCtr = 100000; iCtr--; ) ;	// wait for end of transmission
			ExecuteUserApp(UART_USED);
			break;
		}
	}
}
